/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.core.chunk.rdf;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jactr.core.chunk.IChunk;
import org.jactr.core.chunk.ISymbolicChunk;
import org.jactr.core.chunk.IllegalChunkStateException;
import org.jactr.core.chunk.basic.BasicSymbolicChunk;
import org.jactr.core.chunktype.IChunkType;
import org.jactr.core.module.declarative.IDeclarativeModule;
import org.jactr.core.slot.DefaultMutableSlot;
import org.jactr.core.slot.ISlot;
import org.jactr.core.utils.ModelerException;
import org.openrdf.model.Resource;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.impl.URIImpl;

import eu.larkc.core.data.CloseableIterator;
import eu.larkc.core.data.DataFactory;
import eu.larkc.core.data.LabelledGroupOfStatements;
import eu.larkc.core.data.RdfStoreConnection;
import eu.larkc.jactr.core.chunk.rdf.BasicSymbolicChunkRdf;
import eu.larkc.jactr.core.chunk.rdf.ISymbolicChunkRdf;
import eu.larkc.jactr.modules.declarative.rdf.IDeclarativeModuleRdf;

/**
 * Default Symbolic RDF-chunk implementation
 * 
 * @author Martijn Brakenhoff
 * @created March 15, 2010
 */
public class BasicSymbolicChunkRdf extends BasicSymbolicChunk implements ISymbolicChunkRdf {
	
	private static final transient Log 	LOGGER = LogFactory.getLog(BasicSymbolicChunk.class);
	
	protected static final String 			DEFAULT_CHUNKTYPE			= "chunk";

	protected static final URI RDF_TYPE_URI = new URIImpl("http://www.w3.org/1999/02/22-rdf-syntax-ns#type");
  
	protected 	LabelledGroupOfStatements	_rdf;
	
	protected 	IChunkType					_chunkType;

	protected Resource 						_resource;

	protected IDeclarativeModuleRdf 		_dm;
		
	public BasicSymbolicChunkRdf(IChunk parentChunk, IChunkType ct) {
		super( parentChunk, ct );
		
		IDeclarativeModule dm = ct.getModel().getDeclarativeModule();
		if (dm instanceof IDeclarativeModuleRdf) {
			_dm = (IDeclarativeModuleRdf) dm;
		} else {
			throw new ModelerException("Unable to use RDF Chunks without an IDeclarativeModuleRdf.");
		}
	}

	/**
	 * create a symbolicchunk that has the same values as reference this is used
	 * when making copies of the chunk
	 */
	protected BasicSymbolicChunkRdf(IChunk realParent, ISymbolicChunk reference) {
		this(realParent, reference.getChunkType());
		
		if (reference instanceof ISymbolicChunkRdf) {
			setRdfData( ((ISymbolicChunkRdf) reference).getRdfData() );
		}
	}
	
	public void setName(String name) {
		super.setName(name);
	}

	public LabelledGroupOfStatements getRdfData() {
		try {
		      readLock().lock();
		      return _rdf;
		    } 
			finally {
		      readLock().unlock();
		    }
	}
	
	public void setRdfData(LabelledGroupOfStatements statements) {
		if (_parentChunk.isEncoded()) {
		      throw new IllegalChunkStateException("Cannot change a chunks RDF once it's encoded.");
		}
		
		_rdf = statements;
		
		setSlotsFromRdf( _rdf );
	}
	
	protected void setSlotsFromRdf (LabelledGroupOfStatements rdf) {
		CloseableIterator<Statement> resources = DataFactory.INSTANCE.createRdfStoreConnection().search(null, RDF_TYPE_URI, null, null, rdf.getLabel());
		if (resources.hasNext()) {
			Statement rdfTypeStatement = resources.next();
			
			_resource = rdfTypeStatement.getSubject();
			
			CloseableIterator<Statement> slotsStatements = DataFactory.INSTANCE.createRdfStoreConnection().search(_resource, null, null, null, rdf.getLabel());
			while (slotsStatements.hasNext()) {
				Statement slotStatement = slotsStatements.next();
				
				String slotName = _dm.addPrefix(slotStatement.getPredicate().stringValue());
				String slotValue = _dm.addPrefix(slotStatement.getObject().stringValue());
				
				ISlot ctSlot = getChunkType().getSymbolicChunkType().getSlot(slotName);
				if (ctSlot != null) {
					
					ISlot slot = getSlot(slotName);
					if (slot instanceof DefaultMutableSlot) {
						((DefaultMutableSlot) slot).setValue(slotValue);
					}
					
				}
			}
		}
	}

	@Override
	public void addSlot(ISlot s) {
		  super.addSlot(s);
		  changeValueInRdf(s.getName(), s.getValue());
	}

	protected void changeValueInRdf(String name, Object value) {
		String object = null;
		if (value instanceof String) {
			object = (String) value;
		} else if (value instanceof IChunk) {
			object = ((IChunk) value).getSymbolicChunk().getName();
		}
		
		if (_rdf != null && object != null) {
			RdfStoreConnection con = _dm.createRdfStoreConnection();
			
			String[] nameSplit = name.split(":");
			
			URI predicate = new URIImpl(_dm.getNamespaces().get(nameSplit[0]) + nameSplit[1]);
			
			CloseableIterator<Statement> result = con.search(_resource, predicate, null, null, _rdf.getLabel());
			while (result.hasNext()) {
				Statement s = result.next();				
				con.removeStatement(_resource, predicate, s.getObject(), _rdf.getLabel());
				if (LOGGER.isDebugEnabled()) {
					LOGGER.debug("Removed: " + _resource.stringValue() + " " + predicate.stringValue() + " " + s.getObject().stringValue());
				}
			}
			
			con.addStatement(_resource, predicate, (Value) new URIImpl(object), _rdf.getLabel(), _rdf.getLabel());
			
			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug("Added : " + _resource.stringValue() + " " + predicate.stringValue() + " " + object);
			}
		}
	}

	public Resource getChunkResource() {
		return _resource;
	}
}