/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.core.pluginManager.local;

import java.util.Collection;

import eu.larkc.core.data.InformationSet;
import eu.larkc.core.pluginManager.PluginManager;
import eu.larkc.core.pluginManager.local.LocalPluginManager;
import eu.larkc.core.pluginManager.local.SimpleContract;
import eu.larkc.core.pluginManager.local.queue.Queue;
import eu.larkc.core.query.Query;
import eu.larkc.plugin.Context;
import eu.larkc.plugin.identify.Identifier;

/**
 * LocalSingleResultIdentifyManager is a LocalPluginManager for 
 * Identifier plugins that returns a single result.
 * 
 * It will generate a Collection<InfromationSet> on the output 
 * queue for each Query that comes in on the input stream
 * 
 * @author Martijn Brakenhoff (based on code by: Mick Kerrigan 
 * and Barry Bishop)
 */
public class LocalSingleResultIdentifyManager extends LocalPluginManager <Query, Collection<InformationSet>> {

	/**
	 * The Identifier plugin to be managed
	 */
	private final Identifier mIdentifier;
	
	/**
	 * Constructor thats takes the plugin to be manages, the input, and the output queues as input
	 * 
	 * @param identifier The Identifier plugin to be managed
	 * @param inputQueue The queue from which input messages will come from the previous plugin in the pipeline 
	 * @param outputQueue The queue onto which output messages should be put to send them to the next plugin in the pipeline 
	 */
	public LocalSingleResultIdentifyManager(Identifier identifier, Queue<Query> inputQueue, Queue<Collection<InformationSet>> outputQueue) {
		super(inputQueue, outputQueue);
		mIdentifier = identifier;
		super.setThread(new IdentifierThread());
	}

	/**
	 * The Thread within which the Identifier Management occurs
	 */
	class IdentifierThread extends Thread {
		IdentifierThread() {
			super("Identifier");
		}

		public void run() {
			mIdentifier.initialise();
			Context context = mIdentifier.createContext();
			for (;;) {
				PluginManager.Message controlMessage = getNextControlMessage();

				if (controlMessage.equals(PluginManager.Message.NEXT)) {
					alertPrevious();
					
					Query query = getNextInput();
					
					if (query == null) {
						putNextOutput(null);
						break;
					}
					
					Collection<InformationSet> resources = mIdentifier.identify(query, new SimpleContract(), context);
					putNextOutput(resources);
				}
				else if (controlMessage.equals(PluginManager.Message.STOP)) {
					break;
				}
			}
			stopPrevious();
		}
	}
}
