/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.modules.larkc.plugin;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.concurrent.Future;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jactr.core.buffer.IActivationBuffer;
import org.jactr.core.chunk.IChunk;
import org.jactr.core.chunktype.IChunkType;
import org.jactr.core.model.IModel;
import org.jactr.core.module.AbstractModule;
import org.jactr.core.module.declarative.IDeclarativeModule;
import org.jactr.core.module.retrieval.buffer.DefaultRetrievalBuffer6;
import org.jactr.core.module.retrieval.event.IRetrievalModuleListener;
import org.jactr.core.module.retrieval.four.IRetrievalModule4;
import org.jactr.core.module.retrieval.time.DefaultRetrievalTimeEquation;
import org.jactr.core.module.retrieval.time.IRetrievalTimeEquation;
import org.jactr.core.production.request.ChunkTypeRequest;
import org.jactr.core.slot.ISlot;
import org.jactr.core.utils.ModelerException;
import org.jactr.core.utils.parameter.IParameterized;
import org.jactr.core.utils.parameter.ParameterHandler;

import eu.larkc.core.data.InformationSet;
import eu.larkc.core.data.SetOfStatements;
import eu.larkc.core.pluginManager.PluginManager;
import eu.larkc.core.pluginManager.local.LocalCollectionInformationSetTransformManager;
import eu.larkc.core.pluginManager.local.LocalIdentifyManager;
import eu.larkc.core.pluginManager.local.LocalQueryTransformManager;
import eu.larkc.core.pluginManager.local.LocalReasonManagerFromSetOfStatementsManager;
import eu.larkc.core.pluginManager.local.LocalSelectFromCollectionInformationSetManager;
import eu.larkc.core.pluginManager.local.queue.Queue;
import eu.larkc.core.query.Query;
import eu.larkc.jactr.core.chunk.larkc.IChunkLarKC;
import eu.larkc.jactr.core.chunk.larkc.ISymbolicChunkLarKC;
import eu.larkc.jactr.modules.declarative.rdf.IDeclarativeModuleRdf;
import eu.larkc.jactr.modules.sparql.ISPARQLRetrievalModule;
import eu.larkc.jactr.modules.sparql.buffer.DefaultSPARQLRetrievalBuffer;
import eu.larkc.jactr.scripting.sparql.SPARQLRequest;
import eu.larkc.plugin.Plugin;
import eu.larkc.plugin.identify.Identifier;
import eu.larkc.plugin.reason.Reasoner;
import eu.larkc.plugin.select.Selecter;
import eu.larkc.plugin.transform.InformationSetTransformer;
import eu.larkc.plugin.transform.QueryTransformer;

/**
 * The Module that wraps any LarKC Plug-in. Should be 
 * initialized with the parameters: PluginURI and 
 * PluginBuffer.
 * 
 * @author Martijn Brakenhoff
 * @created April 2, 2010
 */
public class LarKCPluginModule extends AbstractModule implements ILarKCPluginModule, 
ISPARQLRetrievalModule, IRetrievalModule4, IParameterized {
 
	private static final transient Log 	LOGGER = LogFactory.getLog(LarKCPluginModule.class);
	protected static final String		MODULE_NAME = "LarKCPlugin-",
										DEFAULT_BUFFER_NAME	= "LarKC",
										LARKC_DATA_CHUNKTYPE_NAME = "LarKCData",
										LARKC_CHUNK_PREFIX 	= "larkc-",
										RDF_SLOT_NAME		= "rdf";
	protected static int 				numberOfInstances = 0,
										numberOfLarKCChunks = 0;
	
	protected DefaultRetrievalBuffer6	retrievalBuffer;
	protected DefaultSPARQLRetrievalBuffer queryBuffer;
	protected IChunkType 				larkcDataChunkType;
	protected boolean					bufferLocked = false;
	private double						retrievalThreshold = Double.NEGATIVE_INFINITY,
										latencyFactor 		= 20,
										latencyExponent 	= 1;
	private String						bufferName,
										pluginURI;
	private IRetrievalTimeEquation		retrievalTimeEquation;
	private Query						pluginQuery;
	@SuppressWarnings("rawtypes")
	private Queue						pluginInputQueue, 
										pluginOutputQueue;
	private PluginManager				pluginManager;
	private boolean 					isPluginManagerAlive = true;
	private Plugin 						plugin;
	private Class<?> 					pluginInputType,
										pluginOutputType;

	public LarKCPluginModule() {
		this(MODULE_NAME + numberOfInstances);
		numberOfInstances++;
	}

	public LarKCPluginModule(String moduleName) {
		super(moduleName);

		retrievalTimeEquation = new DefaultRetrievalTimeEquation(this);
	}

	// +-------------------+
	// | Module Management |
	// +-------------------+

	@Override
	protected Collection<IActivationBuffer> createBuffers() {
		Collection<IActivationBuffer> result = new HashSet<IActivationBuffer>();

		setBufferName(DEFAULT_BUFFER_NAME);

		retrievalBuffer = new DefaultRetrievalBuffer6( getBufferName(), this );
		result.add(retrievalBuffer);

		if (isPluginAReasoner()) {
			queryBuffer = new DefaultSPARQLRetrievalBuffer( getBufferName() + "Query", this );
			result.add(queryBuffer);
		}
		
		if (LOGGER.isDebugEnabled()) {
			for (IActivationBuffer buffer : result) {
				LOGGER.debug("Added buffer: "+ buffer.getName());
			}
		}
		
		return Collections.unmodifiableCollection(result);
	}

	private boolean isPluginAReasoner() {
		try {
			return Reasoner.class.isAssignableFrom(Class.forName(getPluginURI()));
		} catch (ClassNotFoundException e) {
			return false;
		}
	}

	@Override
	public void install(IModel model) {
		super.install(model);

		setPlugin(instantiatePlugin());
		
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Instantiated Plug-in of type: "+ getPlugin().getClass().getSimpleName());
		}
		
		installPlugin();
		startPluginManager();
	}

	protected Plugin instantiatePlugin() {
		try {
			return (Plugin) Class.forName(getPluginURI()).newInstance();
		} catch (Exception e) {
			if (LOGGER.isFatalEnabled()) { 
				LOGGER.fatal("Exception while Instantiating Plug-in", e); 
			}
			e.printStackTrace();
		}
		
		throw new ModelerException("Unable to instantiate Plug-in '"+ getPluginURI() +"'");
	}

	@Override
	public void initialize() {
		try {
			larkcDataChunkType = getModel().getDeclarativeModule().getChunkType(LARKC_DATA_CHUNKTYPE_NAME).get();
		} catch (InterruptedException e) {
			if (LOGGER.isWarnEnabled()) { 
				LOGGER.warn("Unable to find LarKCData chunktype.", e); 
			}
			e.printStackTrace();
		} catch (ExecutionException e) {
			if (LOGGER.isWarnEnabled()) { 
				LOGGER.warn("Unable to find LarKCData chunktype.", e); 
			}
			e.printStackTrace();
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Initializing "+ getClass().getSimpleName() +" ("+ getBufferName() +") with plugin "+ getPlugin().getClass().getSimpleName());
		}
	}

	@Override
	public void dispose() {
		super.dispose();
		getPluginManager().accept(PluginManager.Message.STOP);
		getPlugin().shutdown();
	}

	@Override
	public void reset() {
		retrievalBuffer.clear();
	}

	@Override
	public void reset(boolean resetFinstsToo) {
		reset();
	}

	public void addListener(IRetrievalModuleListener listener, Executor executor) {
		// TODO Auto-generated method stub
	}

	public void removeListener(IRetrievalModuleListener listener) {
		// TODO Auto-generated method stub
	}

	// +--------------------+
	// | Plug-in Management |
	// +--------------------+

	protected void installPlugin () {
		Plugin plugin = getPlugin(); 
		if (plugin instanceof QueryTransformer) {
			installQueryTransformerPlugin(plugin);
		} else if (plugin instanceof Identifier) {
			installIdentifierPlugin(plugin);
		} else if (plugin instanceof InformationSetTransformer) {
			installInformationSetTransformerPlugin(plugin);
		} else if (plugin instanceof Selecter) {
			installSelecterPlugin(plugin);
		} else if (plugin instanceof Reasoner) {
			installReasonerPlugin(plugin);
		}else {
			throw new IllegalStateException("Unknown plugin tyoe");			
		}
	}

	private void installQueryTransformerPlugin(Plugin plugin) {
		Queue<Query> inputQueue = new Queue<Query>();
		Queue<Query> outputQueue = new Queue<Query>();
		LocalQueryTransformManager manager = new LocalQueryTransformManager((QueryTransformer) plugin, inputQueue, 
				outputQueue);

		setPluginInputType(Query.class);
		setPluginOutputType(Query.class);
		setPluginInputQueue(inputQueue);
		setPluginOutputQueue(outputQueue);
		setPluginManager(manager);
	}

	private void installIdentifierPlugin(Plugin plugin) {
		Queue<Query> inputQueue = new Queue<Query>();
		Queue<Collection<InformationSet>> outputQueue = new Queue<Collection<InformationSet>>();
		LocalIdentifyManager manager = new LocalIdentifyManager((Identifier) plugin, inputQueue, outputQueue);

		setPluginInputType(Query.class);
		setPluginOutputType(Collection.class);
		setPluginInputQueue(inputQueue);
		setPluginOutputQueue(outputQueue);
		setPluginManager(manager);
	}

	private void installInformationSetTransformerPlugin(Plugin plugin) {
		Queue<Collection<InformationSet>> inputQueue = new Queue<Collection<InformationSet>>();
		Queue<Collection<InformationSet>> outputQueue = new Queue<Collection<InformationSet>>();
		LocalCollectionInformationSetTransformManager manager = new LocalCollectionInformationSetTransformManager(
				(InformationSetTransformer) plugin, inputQueue, outputQueue);

		setPluginInputType(Collection.class);
		setPluginOutputType(Collection.class);
		setPluginInputQueue(inputQueue);
		setPluginOutputQueue(outputQueue);
		setPluginManager(manager);
	}

	private void installSelecterPlugin(Plugin plugin) {
		Queue<Collection<InformationSet>> inputQueue = new Queue<Collection<InformationSet>>();
		Queue<SetOfStatements> outputQueue = new Queue<SetOfStatements>();
		LocalSelectFromCollectionInformationSetManager manager = new LocalSelectFromCollectionInformationSetManager(
				(Selecter)plugin, inputQueue, outputQueue);

		setPluginInputType(Collection.class);
		setPluginOutputType(SetOfStatements.class);
		setPluginInputQueue(inputQueue);
		setPluginOutputQueue(outputQueue);
		setPluginManager(manager);
	}

	private void installReasonerPlugin(Plugin plugin) {
		Queue<SetOfStatements> inputQueue = new Queue<SetOfStatements>();
		Queue<InformationSet> outputQueue = new Queue<InformationSet>();
		final LocalReasonManagerFromSetOfStatementsManager manager = new LocalReasonManagerFromSetOfStatementsManager(
				(Reasoner)plugin, getPluginQuery(), inputQueue, outputQueue);
		
		setPluginInputType(SetOfStatements.class);
		setPluginOutputType(InformationSet.class);
		setPluginInputQueue(inputQueue);
		setPluginOutputQueue(outputQueue);
		setPluginManager(manager);
	}

	protected void startPluginManager() {
		if (getPluginManager() != null && isAlive()) {
			getPluginManager().start();
		}
	}

	/**
	 * Returns the Plug-in. Should not be called before installation.
	 * 
	 * @return the Plug-in.
	 */
	protected Plugin getPlugin() {
		return plugin;
	}

	private void setPlugin(Plugin plugin) {
		this.plugin = plugin;
	}

	
	@SuppressWarnings("unchecked")
	protected Queue<Object> getPluginInputQueue() {
		return pluginInputQueue;
	}

	protected void setPluginInputQueue(Queue<?> inputQueue) {
		pluginInputQueue = inputQueue;
	}

	protected Queue<?> getPluginOutputQueue() {
		return pluginOutputQueue;
	}

	protected void setPluginOutputQueue(Queue<?> outputQueue) {
		pluginOutputQueue = outputQueue;
	}

	protected Query getPluginQuery() {
		return pluginQuery;
	}

	protected void setPluginQuery(Query query) {
		this.pluginQuery = query;
	}

	protected PluginManager getPluginManager() {
		return pluginManager;
	}

	protected void setPluginManager(PluginManager pluginManager) {
		this.pluginManager = pluginManager;
	}
	
	protected Class<?> getPluginOutputType() {
		return pluginOutputType;
	}

	protected void setPluginOutputType(Class<?> type) {
		this.pluginOutputType = type;
	}
	
	protected Class<?> getPluginInputType() {
		return pluginInputType;
	}

	protected void setPluginInputType(Class<?> type) {
		this.pluginInputType = type;
	}

	public boolean isAlive() {
		return isPluginManagerAlive;
	}

	private void setAlive(boolean alive) {
		this.isPluginManagerAlive = alive;
	}

	// +--------------------+
	// | Plugin Interaction |
	// +--------------------+

	protected void runPlugin() {
		getPluginManager().accept(PluginManager.Message.NEXT);
	}

	protected void addChunkToPluginInputQueue(ChunkTypeRequest chunkPattern) {
		Collection<ISlot> chunkSlots = chunkPattern.getSlots(new ArrayList<ISlot>());
		for (ISlot slot : chunkSlots) {
			if (slot.getName().equals(RDF_SLOT_NAME)) {
				addObjectToPluginInputQueue(slot.getValue());
				
				return;
			}
		}
		throw new IllegalArgumentException("The given chunk needs to contain LarKC RDF Data.");
	}

	protected void addObjectToPluginInputQueue (Object o) {
		if (o.getClass().equals(getPluginInputType())) {
			throw new IllegalArgumentException("The object provided to the module '" + this.getName() + "' must be compatible with the plug-in it contains.");
		}

		getPluginInputQueue().put(o);
		if (LOGGER.isTraceEnabled()) {
			LOGGER.trace(this.getName() +" added LarKC RDF to the plug-in input queue: "+ o.toString());
		}
	}

	protected Object getNextOutput() {
		Object result = getPluginOutputQueue().take();

		if (result == null) {
			setAlive(false);
		}

		return result;
	}

	// +--------------------+
	// | Retrieval Requests |
	// +--------------------+

	public Future<IChunk> retrieveChunk(final SPARQLRequest sparqlRequest) {
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug(getName() + "("+ getBufferName() +"): Was queried.");
		}
		setPluginQuery(sparqlRequest.getSPARQLQuery());

		installPlugin();
		startPluginManager();

		return delayedFuture(new Callable<IChunk>() {

			public IChunk call() throws Exception {
				// When the chunk get's called, nobody is home.
				return getModel().getDeclarativeModule().getFreeChunk();
			}

		}, getExecutor());
	}

	public Future<IChunk> retrieveChunk(final ChunkTypeRequest chunkPattern) {
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug(getName() + "("+ getBufferName() +"): Got a retrieval request.");
		}

		if (isAlive() && chunkPattern.getChunkType() == larkcDataChunkType) {
			return processByPlugin(chunkPattern);
		} else {
			// Not an active LarKC Plugin or not a valid larkc result chunk, return Error.
			return getFutureErrorChunk();
		}
	}

	protected Future<IChunk> processByPlugin(ChunkTypeRequest chunkPattern) {
		// Fire up the plug-in
		addChunkToPluginInputQueue(chunkPattern);
		runPlugin();

		// Return the promise of a chunk, sometime in the future.
		return delayedFuture(new Callable<IChunk>() {

			public IChunk call() throws Exception {
				// When the chunk get's called look in the output queue
				return putInRdfChunk( getNextOutput() ); 
			}

		}, getExecutor());
	}

	private Future<IChunk> getFutureErrorChunk() {
		return delayedFuture(new Callable<IChunk>() {

			public IChunk call() throws Exception {
				return getModel().getDeclarativeModule().getErrorChunk();
			}
		}, getExecutor());
	}

	protected IChunk putInRdfChunk(Object pluginOutput) {
		IDeclarativeModule dm = getModel().getDeclarativeModule();
		if (! (dm instanceof IDeclarativeModuleRdf) || larkcDataChunkType == null) {
			return dm.getErrorChunk();
		}

		IChunk chunk = null;
		try {
			chunk = dm.createChunk(larkcDataChunkType, LARKC_CHUNK_PREFIX + numberOfLarKCChunks).get();
			numberOfLarKCChunks++;
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (ExecutionException e) {
			e.printStackTrace();
		}

		if (chunk != null && chunk instanceof IChunkLarKC) {

			((IChunkLarKC) chunk).enableLarKCData();
			ISymbolicChunkLarKC sc = (ISymbolicChunkLarKC) chunk.getSymbolicChunk();

			if (pluginOutput instanceof InformationSet) {
				sc.setLarKCData((InformationSet) pluginOutput);
			} else if (pluginOutput instanceof Query) {
				sc.setLarKCData((Query) pluginOutput);
			} else if (pluginOutput instanceof Collection<?>) {
				sc.setLarKCDataCollection((Collection<?>) pluginOutput);
			}

			if (LOGGER.isDebugEnabled()) {
				LOGGER.debug(getName() + "("+ getBufferName() +"): Retrieved result.");
			}
			return chunk;
		} else {
			return dm.getErrorChunk();
		}
	}

	// +--------------------+
	// | Parameter Handling |
	// +--------------------+

	/**
	 * @see org.jactr.core.utils.parameter.IParameterized#getParameter(java.lang.String)
	 */
	public String getParameter(String key) {
		if (RETRIEVAL_THRESHOLD.equalsIgnoreCase(key))	return "" + getRetrievalThreshold();
		if (LATENCY_EXPONENT.equalsIgnoreCase(key))		return "" + getLatencyExponent();
		if (LATENCY_FACTOR.equalsIgnoreCase(key))		return "" + getLatencyFactor();
		return null;
	}

	/**
	 * @see org.jactr.core.utils.parameter.IParameterized#getPossibleParameters()
	 */
	public Collection<String> getPossibleParameters() {
		ArrayList<String> rtn = new ArrayList<String>();
		rtn.add(RETRIEVAL_THRESHOLD);
		rtn.add(LATENCY_EXPONENT);
		rtn.add(LATENCY_FACTOR);
		return rtn;
	}

	/**
	 * @see org.jactr.core.utils.parameter.IParameterized#getSetableParameters()
	 */
	public Collection<String> getSetableParameters() {
		return getPossibleParameters();
	}

	/**
	 * @see org.jactr.core.utils.parameter.IParameterized#setParameter(java.lang.String,
	 *      java.lang.String)
	 */
	public void setParameter(String key, String value) {
		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Module Parameter set: "+ key +"="+ value);
		}

		if (RETRIEVAL_THRESHOLD.equalsIgnoreCase(key)) {
			setRetrievalThreshold(ParameterHandler.numberInstance().coerce(value).doubleValue());
		} else if (LATENCY_EXPONENT.equalsIgnoreCase(key)) {
			setLatencyExponent(ParameterHandler.numberInstance().coerce(value).doubleValue());
		} else if (LATENCY_FACTOR.equalsIgnoreCase(key)) {
			setLatencyFactor(ParameterHandler.numberInstance().coerce(value).doubleValue());
		} else if (BUFFER_NAME.equalsIgnoreCase(key)) {
			setBufferName(value);
		} else if (PLUGIN_URI.equalsIgnoreCase(key)) {
			setPluginURI(value);
		} else if (LOGGER.isWarnEnabled()) {
			LOGGER.warn("No clue how to set " + key + " to " + value);
		}
	}

	public double getRetrievalThreshold () {
		return retrievalThreshold;
	}

	public void setRetrievalThreshold (double threshold) {
		retrievalThreshold = threshold;
	}

	public IRetrievalTimeEquation getRetrievalTimeEquation () {
		return retrievalTimeEquation;
	}

	public double getLatencyExponent () {
		return latencyExponent;
	}

	public void setLatencyExponent (double exp) {
		latencyExponent = exp;
	}

	public double getLatencyFactor () {
		return latencyFactor;
	}

	public void setLatencyFactor (double fact) {
		latencyFactor = fact;
	}	

	public String getPluginURI () {
		return pluginURI;
	}

	public void setPluginURI (String uri) {
		pluginURI = uri;
	}

	public String getBufferName () {
		return bufferName;
	}
	
	public void setBufferName (String name) {
		if (!bufferLocked) {
			bufferLocked = true;
			bufferName = name;
		}
	}
}