/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.modules.sparql;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jactr.core.chunk.IChunk;
import org.jactr.core.chunktype.IChunkType;
import org.jactr.core.chunktype.ISymbolicChunkType;
import org.jactr.core.module.declarative.IDeclarativeModule;
import org.jactr.core.module.retrieval.six.DefaultRetrievalModule6;
import org.jactr.core.production.request.ChunkTypeRequest;
import org.jactr.core.slot.ISlot;
import org.jactr.core.utils.ModelerException;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.impl.URIImpl;

import eu.larkc.core.data.CloseableIterator;
import eu.larkc.core.data.RdfStoreConnection;
import eu.larkc.core.data.SPARQLEndpoint;
import eu.larkc.core.query.RemoteSPARQLEndpoint;
import eu.larkc.core.query.SPARQLQuery;
import eu.larkc.core.query.SPARQLQueryImpl;
import eu.larkc.jactr.modules.declarative.rdf.IDeclarativeModuleRdf;
import eu.larkc.jactr.scripting.sparql.SPARQLRequest;

/**
 * A retrieval module that can retrieve chunks from a repository.
 * 
 * @author Martijn Brakenhoff
 * @created April 27, 2010
 */
public abstract class AbstractSPARQLRetrievalModule extends DefaultRetrievalModule6
implements ISPARQLRetrievalModule {

	private static final Log		LOGGER              = LogFactory.getLog(AbstractSPARQLRetrievalModule.class);

	protected IDeclarativeModuleRdf _dm;

	protected AbstractSPARQLRetrievalModule() {}

	@Override 
	public void initialize() {
		IDeclarativeModule dm = getModel().getDeclarativeModule();
		if (dm instanceof IDeclarativeModuleRdf) {
			_dm = (IDeclarativeModuleRdf) dm;
		} else {
			throw new ModelerException("Unable to use a SPARQL Retrieval Module without " +
			"an IDeclarativeModuleRdf.");
		}
	}

	/**
	 * Returns the SPARQLEndpoint that is used by this module.
	 * 
	 * @return the SPARQLEndpoint.
	 */
	public abstract SPARQLEndpoint getSPARQLEndpoint();

	public Future<IChunk> retrieveChunk(final SPARQLRequest request) {
		return delayedFuture(new Callable<IChunk>() {

			public IChunk call() throws Exception {
				if (LOGGER.isDebugEnabled()){
					LOGGER.debug("Chunk Called");
				}
				return retrieveChunkRdfInternal(request.getSPARQLQuery());
			}

		}, getExecutor());
	}

	protected IChunk retrieveChunkRdfInternal (SPARQLQuery query) {
		SPARQLEndpoint repository = getSPARQLEndpoint();
		CloseableIterator<Statement> queryResult = repository.executeConstruct(query).getStatements();

		if (repository instanceof RemoteSPARQLEndpoint) {
			RdfStoreConnection localStore = _dm.createRdfStoreConnection();
			URI label = new URIImpl("http://larkc.eu/jactr/fromRemote"); 
			while (queryResult.hasNext()) {
				Statement s = queryResult.next();
				localStore.addStatement(s.getSubject(), s.getPredicate(), s
						.getObject(), label, label);
			}
		}

		Collection<IChunk> result = null;
		try {
			result = _dm.findSPARQLMatches(query).get();
		} catch (Exception e) {
			if (LOGGER.isErrorEnabled()){
				LOGGER.error("Threading error while retrieving the SPARQL-Results.", e);
			}
		}

		if (result != null && !result.isEmpty()) {
			IChunk firstChunk = result.iterator().next();
			if (LOGGER.isDebugEnabled()){
				LOGGER.debug("Returning Chunk: "+ firstChunk);
			}
			return firstChunk;
		}
		
		if (LOGGER.isDebugEnabled()){
			LOGGER.debug("Returning Error Chunk");
		}
		
		return _dm.getErrorChunk();
	}

	@Override
	public Future<IChunk> retrieveChunk(final ChunkTypeRequest chunkPattern) {
		return delayedFuture(new Callable<IChunk>() {

			public IChunk call() throws Exception {

				IChunk result = retrieveChunkInternal(getModel().getDeclarativeModule(), chunkPattern);

				if (!result.equalsSymbolic(getModel().getDeclarativeModule().getErrorChunk())) {
					return result;
				} else {
					return retrieveChunk(createSPARQLRequest(chunkPattern)).get();
				}
			}

		}, getExecutor());
	}

	protected SPARQLRequest createSPARQLRequest(ChunkTypeRequest chunkPattern) {
		String query = createSPARQLQuery(chunkPattern.getChunkType());

		return new SPARQLRequest(new SPARQLQueryImpl(query));
	}

	private String createSPARQLQuery(IChunkType chunkType) {
		ISymbolicChunkType sct = chunkType.getSymbolicChunkType();
		String subQueryConstruct = "    ?chunkName rdf:type "+ sct.getName()  +" . \n";
		String subQueryWhere = subQueryConstruct;

		int counter = 0;

		Iterator<ISlot> slots = sct.getSlots(new ArrayList<ISlot>()).iterator();
		while( slots.hasNext() ) {
			ISlot slot = slots.next();

			String triple = "?chunkName "+ slot.getName() +" ?value"+ counter;
			subQueryConstruct	+= "    "+ triple +" . \n";
			subQueryWhere		+= "    OPTIONAL { "+ triple +" } . \n";
			counter++;
		}

		return  "CONSTRUCT { \n" + 
		subQueryConstruct +
		"} WHERE { \n" +
		subQueryWhere +
		"}";
	}

}
