/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.modules.sparql.buffer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jactr.core.buffer.IActivationBuffer;
import org.jactr.core.logging.Logger;
import org.jactr.core.model.IModel;
import org.jactr.core.module.retrieval.IRetrievalModule;
import org.jactr.core.module.retrieval.buffer.RetrievalRequestDelegate;
import org.jactr.core.production.request.ChunkRequest;
import org.jactr.core.production.request.ChunkTypeRequest;
import org.jactr.core.production.request.IRequest;
import org.jactr.core.queue.ITimedEvent;

import eu.larkc.jactr.modules.sparql.ISPARQLRetrievalModule;
import eu.larkc.jactr.modules.sparql.buffer.SPARQLRetrievalRequestDelegate;
import eu.larkc.jactr.scripting.sparql.SPARQLRequest;

/**
 * A jACT-R Request Delegate that is used by the DefaultSPARQLRetrieval.
 * 
 * @author Martijn Brakenhoff
 * @created April 20, 2010
 */
public class SPARQLRetrievalRequestDelegate extends RetrievalRequestDelegate {

	static private final transient Log LOGGER = LogFactory.getLog(SPARQLRetrievalRequestDelegate.class);

	private IRetrievalModule _retrievalModule;

	public SPARQLRetrievalRequestDelegate(IRetrievalModule module) {
		super(module);
		_retrievalModule = module;
	}

	@Override
	public boolean willAccept(IRequest request) {
		return request instanceof SPARQLRequest;
	}

	@Override
	protected boolean isValid(IRequest request, IActivationBuffer buffer) throws IllegalArgumentException {
		return request instanceof SPARQLRequest && ((SPARQLRequest) request).getSPARQLQuery() != null;
	}

	@Override
	protected Object startRequest(IRequest request, IActivationBuffer buffer, double requestTime) {
		IModel model = buffer.getModel();
		/*
		 * abort previous retrieval attempt
		 */
		ITimedEvent previousRetrieval = getCurrentTimedEvent();
		if (previousRetrieval != null && !previousRetrieval.hasAborted() && !previousRetrieval.hasFired()) {
			previousRetrieval.abort();
			/*
			 * log that we are aborting
			 */
			if (Logger.hasLoggers(model) || LOGGER.isDebugEnabled()) {
				String msg = "Aborted retrieval of " + getPreviousRequest();
				if (Logger.hasLoggers(model))
					Logger.log(model, Logger.Stream.RETRIEVAL, msg);
				if (LOGGER.isDebugEnabled()) LOGGER.debug(msg);
			}
		}

		/*
		 * clear the current buffer contents
		 */
		buffer.removeSourceChunk(buffer.getSourceChunk());

		setBusy(buffer);

		/*
		 * indexed retrievals must be enabled, lets just return the chunk
		 */
		if (request instanceof ChunkRequest) {
			return ((ChunkRequest) request).getChunk();
		} 
		if (request instanceof SPARQLRequest) {
			if (_retrievalModule instanceof ISPARQLRetrievalModule) {
				if (LOGGER.isDebugEnabled()){
					LOGGER.debug("Performing a SPARQLRequest on module '"+ _retrievalModule.getName() +"' of class '"+ _retrievalModule.getClass().getName() +"'.");
				}

				return ((ISPARQLRetrievalModule) _retrievalModule).retrieveChunk((SPARQLRequest) request);
			} else {
				if (LOGGER.isErrorEnabled()){
					LOGGER.error("The module is not an ISPARQLRetrievalModule, unable to perform SPARQLRequest.");
				}
			}
		}

		ChunkTypeRequest ctRequest = (ChunkTypeRequest) request;
		return _retrievalModule.retrieveChunk(ctRequest);

	}
}
