/*
 * Copyright (C) 2010, Center for Adaptive Behaviour and Cognition (ABC).
 * Developer: Martijn Brakenhoff 
 * 
 * Max Planck Institute for Human Development
 * (Max-Planck-Institut fr Bildungsforschung)
 * Lentzeallee 94, 14195 Berlin
 * Germany 
 */
package eu.larkc.jactr.scripting.sparql;

import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jactr.core.chunk.ISymbolicChunk;
import org.jactr.core.model.IModel;
import org.jactr.core.module.declarative.IDeclarativeModule;
import org.jactr.core.production.VariableBindings;
import org.jactr.core.production.condition.CannotMatchException;
import org.jactr.core.utils.ModelerException;
import org.jactr.scripting.IScriptableFactory;
import org.jactr.scripting.ScriptSupport;
import org.jactr.scripting.condition.IConditionScript;
import org.openrdf.model.Value;

import eu.larkc.core.data.CloseableIterator;
import eu.larkc.core.data.LabelledGroupOfStatements;
import eu.larkc.core.data.VariableBinding;
import eu.larkc.core.data.VariableBinding.Binding;
import eu.larkc.core.data.util.SPARQLQueryExecutor;
import eu.larkc.core.query.SPARQLQuery;
import eu.larkc.core.query.SPARQLQueryImpl;
import eu.larkc.jactr.core.chunk.rdf.IChunkRdf;
import eu.larkc.jactr.core.chunk.rdf.ISymbolicChunkRdf;
import eu.larkc.jactr.modules.declarative.rdf.IDeclarativeModuleRdf;

/**
 * Scriptable Condition that represents a SPARQL Condition.
 * 
 * @author Martijn Brakenhoff
 * @created March 25, 2010
 */
public class SPARQLCondition implements IConditionScript {
  /**
   * Logger definition
   */
	static private final transient Log 		LOGGER           = LogFactory.getLog(SPARQLCondition.class);

	private final IScriptableFactory		_factory;

	private final String					_script;
  
	private final SPARQLQueryInstantiation	_completedQuery;
	
	public SPARQLCondition(String script, IScriptableFactory factory) {
	    _factory = factory;
	    _script = script;
	    
	    _completedQuery = new SPARQLQueryInstantiation(script, "SELECT");
	}

	public int bind(ScriptSupport scriptSupport, IModel model,
			VariableBindings variableBindings, boolean isIterative)
			throws CannotMatchException {
		Map<String, String> namespaces = getNamespaces(model);
		
		if (namespaces != null) {
			_completedQuery.addPrefixes(namespaces);
		}
		
		SPARQLQuery compiledQuery = null;
	    try {
	    	
	    	compiledQuery = new SPARQLQueryImpl(_completedQuery.getBoundQuery(variableBindings));
	    	
	    } catch (Exception ioe) {
	    	// Any Exception from the implementation means it wasn't a valid SPARQL Query.
	    	LOGGER.error("Error in scriptable condition", ioe);
	    	throw new ModelerException("Error in SPARQL Syntax of Scriptable Condition: ", ioe,
	    			"double check your SPARQL sytanx. The SPARQL parser detected an error.");
	    }
	    
	    VariableBinding queryResult = executeAgainstBufferChunks (variableBindings, compiledQuery );
    	// could return multiple bindings, due to cardinality of relations.
    	
    	if (queryResult != null) {
    		List<String> variables = queryResult.getVariables();
    		CloseableIterator<Binding> results = queryResult.iterator();
    		
    		if (results.hasNext()) {
				List<Value> values = results.next().getValues();
				
				if (values != null) {
					for (int i=0; i < variables.size(); i++) {
						if (LOGGER.isTraceEnabled()) {
							LOGGER.trace("Bound variables: "+ variables.get(i) +"="+ values.get(i).stringValue());
						}
						//scriptSupport.set("=" + variables.get(i).substring(1), values.get(i).stringValue());
						variableBindings.bind("=" + variables.get(i).substring(1), values.get(i).stringValue());
					}
					
					// A match was made and all left-over result-variables were bound.
					return 0;
				}
    		}
    		// No match was made. Perhaps with more variables instantiated, it will work.
    		if (!isIterative) {
    			throw new CannotMatchException("Matched against chunk and returned nothing.");
    		}
    		return 1;
    	} else {
    		// No match was made. Perhaps with more variables instantiated, it will work.
    		if (!isIterative) {
    			throw new CannotMatchException("Matched against chunk and returned nothing.");
    		}
    		return 1; 
    	}
  	}
	
	private Map<String, String> getNamespaces(IModel model) {
		IDeclarativeModule dm = model.getDeclarativeModule();
		if (dm instanceof IDeclarativeModuleRdf) {
			return ((IDeclarativeModuleRdf) dm).getNamespaces();
		}
		throw new ModelerException("Error in Scriptable ICondition", "Unable to perform " +
				"SPARQL actions if the Declarative Module does not support RDF.");
	}

	private VariableBinding executeAgainstBufferChunks ( VariableBindings variableBindings, SPARQLQuery query ) throws CannotMatchException {
		Object sourceChunk = variableBindings.get( "="+ _completedQuery.getBufferName() );
		
		if ( sourceChunk != null && sourceChunk instanceof IChunkRdf ) {
			ISymbolicChunk sc = ((IChunkRdf) sourceChunk).getSymbolicChunk();
			if (sc instanceof ISymbolicChunkRdf) {
				LabelledGroupOfStatements rdf = ((ISymbolicChunkRdf) sc).getRdfData();
				if (rdf != null) {
					return new SPARQLQueryExecutor().executeSelect( query, rdf );
				}
			}
		}
		
		throw new CannotMatchException("Chunk in buffer '" + _completedQuery.getBufferName() + "' cannot be queried. "); 
	}
	
	public IScriptableFactory getFactory() {
		return _factory;
	}
	  
	public String getScript() {
	    return _script;
	}
	  
	public IConditionScript clone(IModel model,
			VariableBindings variableBindings) throws CannotMatchException {
		// TODO Auto-generated method stub
		return this;
	}
	  
	public void dispose() {
	    // noop since we reuse this object in clone
	}
}
